"""NMEA helper utilities for location data from commercial GNSS devices."""

import logging
import re
from copy import deepcopy
from dataclasses import dataclass, asdict
from enum import Enum, IntEnum
from typing import Optional

from fieldedge_utilities.logger import verbose_logging
from fieldedge_utilities.properties import camel_case
from fieldedge_utilities.timestamp import iso_to_ts, ts_to_iso

__all__ = ['GnssFixType', 'GnssFixQuality', 'GnssLocation',
           'validate_nmea', 'parse_nmea_to_location']

_log = logging.getLogger(__name__)


class GnssFixType(IntEnum):
    """Enumerated fix type from NMEA-0183 standard."""
    FIX_NONE = 1
    FIX_2D = 2
    FIX_3D = 3


class GnssFixQuality(IntEnum):
    """Enumerated fix type from NMEA-0183 standard."""
    INVALID = 0
    GPS_SPS = 1
    DGPS = 2
    PPS = 3
    RTK = 4
    FLOAT_RTK = 5
    EST_DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass
class GnssLocation:
    """A location class."""
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    altitude: Optional[float] = None
    speed: Optional[float] = None
    heading: Optional[float] = None
    hdop: Optional[float] = None
    pdop: Optional[float] = None
    vdop: Optional[float] = None
    satellites: Optional[int] = None
    timestamp: Optional[int] = None
    fix_type: Optional[GnssFixType] = None
    fix_quality: Optional[GnssFixQuality] = None

    @property
    def iso_time(self) -> 'str|None':
        if self.timestamp is None:
            return None
        return ts_to_iso(self.timestamp)
    
    def json_compatible(self, **kwargs) -> str:
        lat_lon_precision = kwargs.get('lat_lon_precision', 5)
        other_precision = kwargs.get('other_precision', 1)
        result = { k: v for k, v in asdict(self).items() if v is not None }
        if self.timestamp is not None:
            result['iso_time'] = self.iso_time
        for k, v in result.items():
            if k in ['latitude', 'longitude']:
                result[k] = round(v, lat_lon_precision)
            elif isinstance(k, float):
                result[k] = round(v, other_precision)
            elif isinstance(v, Enum):
                result[k] = v.name
        return { camel_case(k): v for k, v in result.items() if v is not None }


def validate_nmea(nmea_sentence: str) -> bool:
    """Validates a given NMEA-0183 sentence with CRC.
    
    Args:
        nmea_sentence (str): NMEA-0183 sentence ending in checksum.
    
    """
    if '*' not in nmea_sentence:
        return False
    data, cs_hex = nmea_sentence.split('*')
    candidate = int(cs_hex, 16)
    crc = 0   # initial
    for i in range(1, len(data)):   # ignore initial $
        crc ^= ord(data[i])
    return candidate == crc


def parse_nmea_to_location(nmea_sentence: str,
                           location: GnssLocation = None,
                           **kwargs) -> 'dict|GnssLocation|None':
    """Parses a NMEA-0183 sentence to a location or update.
    
    Passing a Location object in will update the location with NMEA data.
    Otherwise a dictionary is returned.
    """
    if _vlog():
        _log.debug('Parsing NMEA: %s', nmea_sentence)
        _log.debug('Previous location: %s', location)
    if not validate_nmea(nmea_sentence):
        raise ValueError('Invalid NMEA-0183 sentence')
    if not isinstance(location, GnssLocation):
        location = GnssLocation()
        old_location = None
    else:
        old_location = deepcopy(location)
    void = False
    data = nmea_sentence.split('*')[0]
    nmea_type = ''
    cache = {}
    for i, field_data in enumerate(data.split(',')):
        if not field_data:
            continue   # skip empty fields
        if i == 0:
            nmea_type = field_data[-3:]
            if nmea_type not in ['RMC', 'GGA', 'GSA']:
                if _vlog():
                    _log.warning('No processing defined for %s sentence',
                                 nmea_type)
                break
            if _vlog():
                _log.debug('Processing NMEA type: %s', nmea_type)
        elif i == 1:
            if nmea_type == 'RMC':
                cache['fix_hour'] = field_data[0:2]
                cache['fix_min'] = field_data[2:4]
                cache['fix_sec'] = field_data[4:6]
                if _vlog():
                    _log.debug('Fix time %s:%s:%s', cache['fix_hour'],
                               cache['fix_min'], cache['fix_sec'])
        elif i == 2:
            if nmea_type == 'RMC':
                if (field_data == 'V'):
                    _log.warning('Fix Void')
                    void = True
                    break
            elif nmea_type == 'GSA':
                location.fix_type = GnssFixType(int(field_data))
                if _vlog():
                    _log.debug('Fix type: %D', location.fix_type.name)
        elif i == 3:
            if nmea_type == 'RMC':
                location.latitude = (float(field_data[0:2]) +
                                     float(field_data[2]) / 60.0)
        elif i == 4:
            if nmea_type == 'RMC':
                if field_data == 'S':
                    location.latitude *= -1
                if _vlog():
                    _log.debug('Latitude: %.5f', location.latitude)
        elif i == 5:
            if nmea_type == 'RMC':
                location.longitude = (float(field_data[0:3]) +
                                      float(field_data[3]) / 60.0)
        elif i == 6:
            if nmea_type == 'RMC':
                if field_data == 'W':
                    location.longitude *= -1
                if _vlog():
                    _log.debug('Longitude: %.5f', location.longitude)
            elif nmea_type == 'GGA':
                location.fix_quality = GnssFixQuality(int(field_data))
                if _vlog():
                    _log.debug('Fix quality: %s', location.fix_quality.name)
        elif i == 7:
            if nmea_type == 'RMC':
                location.speed = float(field_data) * 1.852
                if _vlog():
                    _log.debug('Speed: %.1f', location.speed)
            elif nmea_type == 'GGA':
                location.satellites = int(field_data)
                if _vlog():
                    _log.debug('GNSS satellites used: %d', location.satellites)
        elif i == 8:
            if nmea_type == 'RMC':
                location.heading = float(field_data)
                if _vlog():
                    _log.debug('Heading: %.1f', location.heading)
            elif nmea_type == 'GGA':
                location.hdop = round(float(field_data), 1)
                if _vlog():
                    _log.debug('HDOP: %.1f', location.hdop)
        elif i == 9:
            if nmea_type == 'RMC':
                fix_day = field_data[0:2]
                fix_month = field_data[2:4]
                fix_yy = int(field_data[4:])
                fix_yy += 1900 if fix_yy >= 73 else 2000
                if _vlog():
                    _log.debug('Fix date %d-%s-%s', fix_yy, fix_month, fix_day)
                iso_time = (f'{fix_yy}-{fix_month}-{fix_day}T'
                            f'{cache["fix_hour"]}:{cache["fix_min"]}'
                            f':{cache["fix_sec"]}Z')
                unix_timestamp = iso_to_ts(iso_time)
                if _vlog():
                    _log.debug('Fix time ISO 8601: %s | Unix: %d',
                               iso_time, unix_timestamp)
                location.timestamp = unix_timestamp
            elif nmea_type == 'GGA':
                location.altitude = float(field_data)
                if _vlog():
                    _log.debug('Altitude: %.1f', location.altitude)
        elif i == 10:
            # RMC magnetic variation - ignore
            if nmea_type == 'GGA' and field_data != 'M':
                _log.warning('Unexpected altitude units: %s', field_data)
        # elif i == 11:   # RMC magnetic variation direction, GGA height of geoid - ignore
        # elif i == 12:   # GGA units height of geoid - ignore
        # elif i == 13:   # GGA seconds since last DGPS update - ignore
        # elif i == 14:   # GGA DGPS station ID - ignore
        elif i == 15:   # GSA PDOP - ignore (unused)
            if nmea_type == 'GSA':
                location.pdop = round(float(field_data), 1)
                if _vlog():
                    _log.debug('PDOP: %d', location.pdop)
        # elif i == 16:   # GSA HDOP - ignore (use GGA)
        elif i == 17:
            if nmea_type == 'GSA':
                location.vdop = round(float(field_data), 1)
                if _vlog():
                    _log.debug('VDOP: %d', location.vdop)
    if void:
         if isinstance(old_location, GnssLocation):
             return old_location
         return None
    if isinstance(old_location, GnssLocation):
        return location
    return location.json_compatible(**kwargs)


def _vlog() -> bool:
    """Check if vebose logging is enabled for this microservice."""
    return verbose_logging('gnss')


# Placeholder regex solution.  Regex broken for GSA
# rmc_pattern = re.compile(
#     r'^\$G[NP]RMC,'                # Match $GPRMC or $GNRMC (GPS or multi-GNSS)
#     r'(\d{6}(?:\.\d+)?)'           # UTC Time (hhmmss.sss)
#     r',([AV])'                     # Status (A = Valid, V = Void)
#     r',(\d{2,3}\d*\.\d+),([NS])'   # Latitude (ddmm.mmmm), N/S
#     r',(\d{3,4}\d*\.\d+),([EW])'   # Longitude (dddmm.mmmm), E/W
#     r',(\d+\.\d+)?'                # Speed over ground (knots)
#     r',(\d+\.\d+)?'                # Course over ground (degrees)
#     r',(\d{6})'                    # Date (DDMMYY)
#     r'(,-?\d+\.\d+)?,([EW])?'      # Magnetic variation (optional), E/W
#     r'(,[ADEN]?)?\*[\dA-Fa-f]{2}$' # Optional mode indicator & checksum
# )
# gga_pattern = re.compile(
#     r'^\$G[NP]GGA,'                      # Match $GPGGA or $GNGGA
#     r'(\d{6}(?:\.\d+)?)'           # UTC Time (hhmmss.sss)
#     r',(\d{2,3}\d*\.\d+),([NS])'   # Latitude (ddmm.mmmm), N/S
#     r',(\d{3,4}\d*\.\d+),([EW])'   # Longitude (dddmm.mmmm), E/W
#     r',(\d)'                              # Fix quality (0-8)
#     r',(\d{1,2})'                             # Number of satellites
#     r',([\d\.]+)'                         # HDOP
#     r',(-?[\d\.]+),M'                     # Altitude (meters) + M
#     r',(-?[\d\.]+),M'                     # Geoidal separation (meters) + M
#     r',(?:,([\d\.]+),)?'                    # DGPS Age (optional)
#     r'(?:,(\d+))?'                        # DGPS Station ID (optional)
#     r'\*([\dA-Fa-f]{2})$'                 # Checksum
# )
# gsa_pattern = re.compile(
#     r'^\$G[NP]GSA,'           # Match $GPGSA or $GNGSA
#     r'([AM]),'                # Mode (A/M)
#     r'([123]),'               # Fix type (1 = No Fix, 2 = 2D, 3 = 3D)
#     r'(\d{2}|,){12}'          # Broken
#     r'([\d\.]+),'              # PDOP (1.8)
#     r'([\d\.]+),'              # HDOP (1.0)
#     r'([\d.]+)'               # VDOP (1.5)
#     r'\*([\dA-Fa-f]{2})$'     # Checksum (*3E)
# )


# def parse_nmea_sentence(nmea) -> dict:
#     """Parse an NMEA sentence and extract location-based information.

#     Args:
#         nmea (str): A NMEA-0183 sentence

#     Returns:
#         `dictionary` with keys derived from NMEA like `latitude`, `longitude`,
#             `heading`, `speed`, `fix_quality`, and `utc_time`.
#     """
#     result = {}
#     # Parse GGA data (latitude, longitude, fix quality, UTC time)
#     gga_match = gga_pattern.match(nmea)
#     if gga_match:
#         fields = gga_match.groups()
#         utc_tod = fields[0]
#         result['tod'] = utc_tod
#         lat_ddmm = fields[1]
#         lat_dir = fields[2]
#         abs_lat = round(float(lat_ddmm[:2]) + float(lat_ddmm[2:]) / 60, 6)
#         result['latitude'] = abs_lat if lat_dir == 'N' else -abs_lat
#         lon_dddmm = fields[3]
#         lon_dir = fields[4]
#         abs_lon = round(float(lon_dddmm[:3]) + float(lon_dddmm[3:]) / 60, 6)
#         result['longitude'] = abs_lon if lon_dir == 'E' else -abs_lon
#         result['fix_quality'] = int(fields[5])
#         result['num_satellites'] = int(fields[6])
#         result['hdop'] = float(fields[7])
#         result['altitude'] = float(fields[8])
#         result['geoidal_separation'] = float(fields[9])
#         if fields[10]:
#             result['dgps_age'] = fields[10]
#         if fields[11]:
#             result['dgps_station_id'] = fields[11]
#     # Parse RMC data (latitude, longitude, speed, heading, UTC time)
#     rmc_match = rmc_pattern.match(nmea)
#     if rmc_match:
#         fields = rmc_match.groups()
#         status = fields[2]
#         if status == 'V':
#             _log.warning('NMEA sentence void')
#             return result
#         utc_ddmmyy = fields[8]
#         utc_hhmmss = fields[0]
#         d, m, y = int(utc_ddmmyy[:2]), int(utc_ddmmyy[2:4]), int(utc_ddmmyy[4:])
#         y += 1900 if y >= 73 else 2000
#         hh, mm, ss = int(utc_hhmmss[:2]), int(utc_hhmmss[2:4]), int(utc_hhmmss[4:])
#         result['timestamp'] = iso_to_ts(f'{y}-{m:02}-{d:02}T{hh:02}:{mm:02}:{ss:02}Z')
#         lat_ddmm = fields[2]
#         lat_dir = fields[3]
#         abs_lat = round(float(lat_ddmm[:2]) + float(lat_ddmm[2:]) / 60, 6)
#         result['latitude'] = abs_lat if lat_dir == 'N' else -abs_lat
#         lon_dddmm = fields[4]
#         lon_dir = fields[5]
#         abs_lon = round(float(lon_dddmm[:3]) + float(lon_dddmm[3:]) / 60, 6)
#         result['longitude'] = abs_lon if lon_dir == 'E' else -abs_lon
#         speed_kn = fields[6]
#         result['speed'] = round(float(speed_kn) / 1.852 if speed_kn else 0.0, 1)
#         cog = fields[7]
#         result['course'] = round(float(cog) if cog else 0.0, 1)
#         # mag_var = fields[9] + fields[10]
#     # Parse GSA data (fix type)
#     gsa_match = gsa_pattern.match(nmea)
#     if gsa_match:
#         fields = gsa_match.groups()
#         fix_quality = int(gsa_match.group(1))
#         result['fix_quality'] = fix_quality
#     return result



"""Wrapping logger common format for FieldEdge project(s) with UTC timestamp.

Provides a common log format with a console and file handler.
The file is a wrapping log of configurable size using `RotatingFileHandler`.

Format is:

* ISO UTC timestamp (datetime) e.g. `2021-01-01T00:00:00.000Z`
* Log level, CSV encloses in square brackets e.g. `[INFO]`
* Thread name, CSV encloses in round brackets
* Module, Function and Line. CSV uses `ModuleName.FunctionName:(LineNumber)`
* Message

*CSV Example:*

`2021-10-30T14:19:51.012Z,[INFO],(MainThread),main.<module>:6,This is a test`

*JSON Example:*

`{"timestamp":"2021-01-01T00:00:00Z","level":"INFO","thread":"MainThread",
"module":"main","function":"<module>","line":6,"message":"This is a test"}`

"""
import json
import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from time import gmtime

from fieldedge_utilities.path import clean_path

FORMAT_CSV = ('%(asctime)s.%(msecs)03dZ,[%(levelname)s],(%(threadName)s),'
              '%(module)s.%(funcName)s:%(lineno)d,%(message)s')
FORMAT_JSON = ('{'
                '"datetime":"%(asctime)s.%(msecs)03dZ"'
                ',"level":"%(levelname)s"'
                ',"thread":"%(threadName)s"'
                ',"module":"%(module)s"'
                ',"function":"%(funcName)s"'
                ',"line":%(lineno)d'
                ',"message":"%(message)s"'
                '}')
DATEFMT = '%Y-%m-%dT%H:%M:%S'

DEFAULT_OBSCURE = ['password', 'token', 'key', 'secret']


class LogFilterLessThan(logging.Filter):
    """Filters logs below a specified level for routing to a given handler.
    
    Intended to route `WARNING` and higher to `STDERR` and lower to `STDOUT`.

    """
    def __init__(self,
                 exclusive_maximum: int = logging.WARNING,
                 name: str = None):
        if name is None:
            name = f'LessThan{logging.getLevelName(exclusive_maximum)}'
        super().__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        #non-zero return means we log this message
        return 1 if record.levelno < self.max_level else 0


class LogFormatterOneLineException(logging.Formatter):
    """Formats exceptions into a single line stack trace.
    
    Also replaces the record message with the error type.

    """
    def formatException(self, exc_info):
        original = super().formatException(exc_info)
        return ' -> '.join([x.strip() for x in original.splitlines()])

    def format(self, record):
        result = super().format(record)
        if record.exc_text:
            err_type = type(record.msg).__name__
            result = result.replace(f'{record.msg}\n', f'{err_type}: ')
        return result


class LogFormatterObscureSensitive(LogFormatterOneLineException):
    """Formats all text to remove sensitive information."""
    @staticmethod
    def obscure(record: str) -> str:
        keywords = json.loads(os.getenv('OBSCURE', json.dumps(DEFAULT_OBSCURE)))
        if not any(k in record.lower() for k in keywords):
            return record
        wrappers = ['"', '\'']
        assignments = [': ', ':', '=', ' = ']
        separators = [',', ' ', '}', ']']
        to_replace = []
        for k in keywords:
            if k in record.lower():
                index = 0
                while index < len(record) - 1:
                    tag = k
                    index = record[index:].find(tag)
                    if index == -1:
                        break
                    for w in wrappers:
                        if record[index+len(tag)] == w:
                            tag = tag + w
                            break
                    for a in assignments:
                        if record[index+len(tag):index+len(tag)+len(a)] == a:
                            tag = tag + a
                            break
                    start = index + len(tag)
                    end = len(record) - start
                    for s in separators:
                        end = record[start:].find(s)
                        if end > -1:
                            break
                        end = len(record) - start
                    pattern = record[start:start+end]
                    for w in wrappers:
                        if pattern.startswith(w) and pattern.endswith(w):
                            pattern = pattern[len(w):-len(w)]
                            break
                    if pattern not in to_replace:
                        to_replace.append(pattern)
                    index = start + end + 1
        for r in to_replace:
            record = record.replace(r, '***')
        return record
    
    def format(self, record):
        preformatted = super().format(record)
        return self.obscure(preformatted)


def get_logfile_name(logger: logging.Logger) -> str:
    """Returns the logger's RotatingFileHandler name.
    
    Args:
        logger: The Logger to retrieve its file name from.

    Raises:
        TypeError if Logger is invalid

    """
    if not isinstance(logger, logging.Logger):
        raise TypeError('Invalid Logger')
    for h in logger.handlers:
        if isinstance(h, RotatingFileHandler):
            return h.baseFilename


def add_handler(logger: logging.Logger, handler: logging.Handler) -> None:
    """Adds a handler to a logger if one of the same name is not present.
    
    Args:
        logger: The logger to add the handler to.
        handler: The handler to add to the logger.
    
    Raises:
        TypeError if Logger or Handler are invalid
        ValueError if a handler with the same name is already in the logger.

    """
    if not isinstance(logger, logging.Logger):
        raise TypeError('Invalid Logger')
    if not isinstance(handler, logging.Handler):
        raise TypeError('Invalid Handler')
    for h in logger.handlers:
        if h.name == handler.name:
            raise ValueError(f'Logger already has a handler named {h.name}')
    logger.addHandler(handler)


def apply_formatter(logger: logging.Logger,
                    formatter: logging.Formatter) -> None:
    """Applies the log formatter to all handlers in the logger.
    
    Args:
        logger: The logger with the handler targets.
        formatter: The formatter to apply.

    Raises:
        TypeError if Logger or Handler are invalid
    
    """
    if not isinstance(logger, logging.Logger):
        raise TypeError('Invalid Logger')
    if not isinstance(formatter, logging.Formatter):
        raise TypeError('Invalid Formatter')
    for h in logger.handlers:
        h.setFormatter(formatter)


def apply_loglevel(logger: logging.Logger, level: int) -> None:
    """Sets the log level on all handlers except stderr (always WARNING).
    
    Args:
        logger: The logger whose handlers to apply the level to.
        level: The logging level to apply

    Raises:
        TypeError if Logger is invalid
    
    """
    if not isinstance(logger, logging.Logger):
        raise TypeError('Invalid Logger')
    for h in logger.handlers:
        if 'stderr' not in h.name:
            h.setLevel(level)


def get_formatter(format: str = 'csv',
                  obscure: bool = False) -> logging.Formatter:
    """Returns a standardized log formatter.
    
    Args:
        format: `csv` or `json` are supported.
        obscure: `True` will obscure logged text with keys matching environment
            variable `OBSCURE` as a JSON-formatted array.
    
    Returns:
        A `logging.Formatter`

    """
    fmt = FORMAT_JSON if format == 'json' else FORMAT_CSV
    if obscure:
        log_formatter = LogFormatterObscureSensitive(fmt, DATEFMT)
    else:
        log_formatter = LogFormatterOneLineException(fmt, DATEFMT)
    log_formatter.converter = gmtime
    return log_formatter


def get_handler_file(filename: str,
                     file_size: int = 5,
                     **kwargs) -> RotatingFileHandler:
    if not filename:
        raise ValueError('Missing filename')
    filename = clean_path(filename)
    if not os.path.isdir(os.path.dirname(filename)):
        raise FileNotFoundError('Invalid logfile path'
            f' {os.path.dirname(filename)}')
    handler = RotatingFileHandler(
        filename=filename,
        mode=kwargs.pop('mode', 'a'),
        maxBytes=kwargs.pop('maxBytes', int(file_size * 1024 * 1024)),
        backupCount=kwargs.pop('backupCount', 2),
        encoding=kwargs.pop('encoding', None),
        delay=kwargs.pop('delay', 0),
    )
    handler.name = f'{kwargs.pop("name", __name__)}_file_handler'
    return handler


def get_handler_stdout(**kwargs) -> logging.StreamHandler:
    handler_stdout = logging.StreamHandler(sys.stdout)
    handler_stdout.name = f'{kwargs.pop("name", __name__)}_stdout_handler'
    handler_stdout.addFilter(LogFilterLessThan(logging.WARNING))
    return handler_stdout


def get_handler_stderr(**kwargs) -> logging.StreamHandler:
    handler_stderr = logging.StreamHandler(sys.stderr)
    handler_stderr.name = f'{kwargs.pop("name", __name__)}_stderr_handler'
    handler_stderr.setLevel(logging.WARNING)
    return handler_stderr


def get_wrapping_logger(name: str = None,
                        filename: str = None,
                        file_size: int = 5,
                        log_level: int = logging.INFO,
                        format: str = 'csv',
                        **kwargs) -> logging.Logger:
    """Sets up a wrapping logger that writes to console and optionally a file.

    DEPRECATED in favour of get_fieldede_logger

    * Default logging level is `INFO`
    * Timestamps are UTC ISO 8601 format
    * Initializes logging to stdout/stderr, and optionally a CSV or JSON
    formatted file. Default is CSV.
    * Wraps files at a given `file_size` in MB, with default 2 backups.
    
    CSV format: timestamp,[level],(thread),module.function:line,message

    Args:
        name: Name of the logger (if None, uses name of calling module).
        filename: Name of the file/path if writing to a file.
        file_size: Max size of the file in megabytes, before wrapping.
        log_level: the logging level (default INFO)
        format: `csv` or `json`
        kwargs: Optional overrides for RotatingFileHandler
            mode (str): defaults to `a` (append)
            maxBytes (int): overrides file_size
            backupCount (int): defaults to 2
    
    Returns:
        A `Logger` with console `StreamHandler` and (optional)
            `RotatingFileHandler`.
    
    Raises:
        `FileNotFoundError` if a logfile name is specified with an invalid path.
    
    """
    if not name:
        name = __name__
    logger = logging.getLogger(name)
    if filename is not None:
        filename = clean_path(filename)
        if not os.path.isdir(os.path.dirname(filename)):
            raise FileNotFoundError('Invalid logfile path'
                f' {os.path.dirname(filename)}')
        add_handler(logger, get_handler_file(filename,
                                             name=name,
                                             file_size=file_size,
                                             **kwargs))
    add_handler(logger, get_handler_stdout(name=name))
    add_handler(logger, get_handler_stderr(name=name))
    apply_formatter(logger, get_formatter(format))
    logger.setLevel(log_level)
    return logger


def get_fieldedge_logger(filename: str = None,
                         file_size: int = 5,
                         log_level: 'int|str' = logging.INFO,
                         format: str = 'csv',
                         **kwargs) -> logging.Logger:
    """Sets up a root logger that writes to console and optionally a file.

    * Default logging level is `INFO`
    * Timestamps are UTC ISO 8601 format
    * Initializes logging to stdout/stderr, and optionally a CSV or JSON
    formatted file. Default is CSV.
    * Wraps files at a given `file_size` in MB, with default 2 backups.
    
    Args:
        name: Name of the logger (if None, uses name of calling module).
        filename: Name of the file/path if writing to a file.
        file_size: Max size of the file in megabytes, before wrapping.
        log_level: the logging level (default INFO)
        format: `csv` or `json`
    
    Keyword Args:
        mode (str): writing mode defaults to `a` (append)
        maxBytes (int): overrides file_size
        backupCount (int): defaults to 2
        obscure (bool): Set `True` to obscure values in logs defined by the
            `OBSCURE` environment variable as a JSON array of keywords. Default
            keywords: `password`, `token`, `key`, `secret`.
    
    Returns:
        A `Logger` with console `StreamHandler` and (optional)
            `RotatingFileHandler`.
    
    Raises:
        `FileNotFoundError` if a logfile name is specified with an invalid path.
    
    """
    logger = logging.getLogger()
    filehandler_kwargs = {}
    for k, v in kwargs.items():
        if k in ['mode', 'maxBytes', 'backupCount']:
            filehandler_kwargs[k] = v
    if isinstance(log_level, str):
        log_level = log_level.upper()
    if filename is not None:
        filename = clean_path(filename)
        if not os.path.isdir(os.path.dirname(filename)):
            raise FileNotFoundError('Invalid logfile path'
                                    f' {os.path.dirname(filename)}')
        add_handler(logger, get_handler_file(filename,
                                             file_size=file_size,
                                             **kwargs))
    add_handler(logger, get_handler_stdout())
    add_handler(logger, get_handler_stderr())
    apply_formatter(logger, get_formatter(format, kwargs.get('obscure', False)))
    logger.setLevel(log_level)
    return logger


def verbose_logging(filter: str = '', case_sensitive: bool = True) -> bool:
    """Indicates if verbose logging is configured, with an optional filter.
    
    Args:
        filter: An optional filter e.g. the package+module name
        case_sensitive: Defaults to case sensitive filter.
    
    Returns:
        True if `LOG_VERBOSE` environment variable is set
            and includes the filter.
            
    """
    log_verbose = os.getenv('LOG_VERBOSE')
    if log_verbose:
        if filter:
            if (filter in log_verbose or
                not case_sensitive and filter.upper() in log_verbose.upper()):
                return True
            return False
        return True
    return False



"""Geostationary Satellite class for FieldEdge microservices."""

import math
from enum import Enum

__all__ = ['GeoSatellite', 'geo_azimuth', 'geo_elevation', 'geo_closest']

# Constants
R_EARTH = 6378.137  # Radius of Earth in kilometers
R_GEO = 35786       # Altitude of geostationary orbit in kilometers


class GeoSatellite(Enum):
    """A class defining Viasat/Inmarsat geostationary satellite location.
    
    A shorthand name for the geographic zone is given to a satellite longitude.
    """
    AMER = -98.0   # I4F3 at time of writing
    AORW = -54.0   # I3F5 at time of writing
    APAC = 143.5   # I4F2 at time of writing
    EMEA = 24.9   # I4A4 (Alphasat) at time of writing
    IOE = 83.8   # I6F1 at time of writing
    USCA = -101.3   # Ligado SkyTerra-1 at time of writing


def _validate(sat_lon: float, es_lat: float, es_lon: float, precision: int):
    if any(not isinstance(lon, (float, int)) or lon < -180 or lon > 180
           for lon in (sat_lon, es_lon)):
        raise ValueError('Invalid longitude must be in range +/- 180')
    if not isinstance(es_lat, (float, int)) or es_lat < -90 or es_lat > 90:
        raise ValueError('Invalid latitude must be in range +/- 90')
    if not isinstance(precision, int) or precision not in range(0, 4):
        raise ValueError('Invalid precision must be integer in range 0..3')

    
def geo_azimuth(sat_lon: float, es_lat: float, es_lon: float, precision: int = 0) -> 'int|float':
    """Get the azimuth of a geostationary satellite.
    
    Args:
        sat_lon (float): The longitude of the satellite in degrees (+ East)
        es_lat (float): The earth station latitude in degrees (+ North)
        es_lon (float): The earth station longitude in degrees (+ East)
        precision (int): The decimal rounding of the result
    
    Returns:
        The azimuth to the satellite relative to the earth station in degrees.
    """
    _validate(sat_lon, es_lat, es_lon, precision)
    d_lon = math.radians(es_lon - sat_lon)
    es_az_rad = math.atan(math.tan(d_lon)/math.sin(math.radians(es_lat)))
    es_az = 180 + math.degrees(es_az_rad)
    if es_lat < 0:
        es_az = es_az - 180
    if es_az < 0:
        es_az = es_az + 360
    result = round(es_az, precision)
    return result if precision else int(result)


def geo_elevation(sat_lon: float, es_lat: float, es_lon: float, precision: int = 0) -> 'int|float':
    """Get the elevation of a geostationary satellite.
    
    Args:
        sat_lon (float): The longitude of the satellite in degrees (+ East)
        es_lat (float): The earth station latitude in degrees (+ North)
        es_lon (float): The earth station longitude in degrees (+ East)
        precision (int): The decimal rounding of the result
    
    Returns:
        The elevation to the satellite relative to the earth station in degrees.
    """
    _validate(sat_lon, es_lat, es_lon, precision)
    d_lon = math.radians(es_lon - sat_lon)
    es_lat_rad = math.radians(es_lat)
    r1 = 1 + R_GEO / R_EARTH
    v1 = r1 * math.cos(es_lat_rad) * math.cos(d_lon) - 1
    v2 = r1 * math.sqrt(1 - math.cos(es_lat_rad)**2 * math.cos(d_lon)**2)
    es_el = math.degrees(math.atan(v1/v2))
    result = round(es_el, precision)
    return result if precision else int(result)


def geo_closest(es_lat: float, es_lon: float, exclude: 'list[str]' = []) -> 'GeoSatellite':
    """Get the GeoSatellite closest to the specified location.
    
    Args:
        es_lat (float): The earth station latitude in degrees.
        es_lon (float): The earth station longitude in degrees.
        exclude (list[str]): Optional satellite shorthand names to exclude
    
    Returns:
        `GeoSatellite` closest to the earth station.
    """
    sats = [e for e in GeoSatellite]
    sat_names = [e.name for e in GeoSatellite]
    if not isinstance(es_lat, (float, int)) or es_lat < -90 or es_lat > 90:
        raise ValueError('Invalid latitude must be in range +/- 90')
    if not isinstance(es_lon, (float, int)) or es_lon < -180 or es_lon > 180:
        raise ValueError('Invalid longitude must be in range +/- 180')
    if not isinstance(exclude, list) or not all(x in sat_names for x in exclude):
        raise ValueError(f'Invalid exclusion name must be in ({sat_names})')
    if isinstance(exclude, list) and all(isinstance(x, str) for x in exclude):
        rem = []
        for i, sat in enumerate(sats):
            if sat.name in exclude:
                rem.append(i)
        for i in rem:
            del sats[i]
    closest = min(sats, key=lambda x:abs(x.value - es_lon))
    if (GeoSatellite.AORW.name not in exclude and
        closest == GeoSatellite.AORW):
        if es_lat >= 15 or es_lat <= -45:
            if es_lon < -27:
                closest = GeoSatellite.AMER
            else:
                closest = GeoSatellite.EMEA
    if (GeoSatellite.USCA.name not in exclude and
        closest == GeoSatellite.AMER):
        if (es_lat > 25 and es_lat < 60) and es_lon > -123:
            closest = GeoSatellite.USCA
    return closest



"""MQTT client for local broker inter-service communications.

This MQTT client sets up automatic connection and reconnection intended mainly
for use with a local broker on an edge device e.g. Raspberry Pi.

Reads broker configuration from a local `.env` file or environment variables:

* `MQTT_HOST` the IP address or hostname or container of the broker
* `MQTT_USER` the authentication username for the broker
* `MQTT_PASS` the authentication password for the broker

Typically the `fieldedge-broker` will be a **Mosquitto** service running locally
in a **Docker** container listening on port 1883 for authenticated connections.

"""
import json
import logging
import os
import threading
from atexit import register as on_exit
from enum import IntEnum
from socket import gaierror, timeout  # : Python<3.10 vs TimeoutError
from time import sleep, time
from typing import Any, Callable

from paho.mqtt.client import Client as PahoClient
from paho.mqtt.client import MQTTMessage as PahoMessage

from fieldedge_utilities.logger import verbose_logging
from fieldedge_utilities.properties import json_compatible

__all__ = ['MqttResultCode', 'MqttError', 'MqttClient']

_log = logging.getLogger(__name__)


class MqttResultCode(IntEnum):
    """Eclipse Paho MQTT Error Codes."""
    SUCCESS = 0
    ERR_INCORRECT_PROTOCOL = 1
    ERR_INVALID_CLIENT_ID = 2
    ERR_SERVER_UNAVAILABLE = 3
    ERR_BAD_USERNAME_OR_PASSWORD = 4
    ERR_UNAUTHORIZED = 5
    ERR_CONNECTION_LOST = 6
    ERR_TIMEOUT_LENGTH = 7
    ERR_TIMEOUT_PAYLOAD = 8
    ERR_TIMEOUT_CONNACK = 9
    ERR_TIMEOUT_SUBACK = 10
    ERR_TIMEOUT_UNSUBACK = 11
    ERR_TIMEOUT_PINGRESP = 12
    ERR_MALFORMED_LENGTH = 13
    ERR_COMMUNICATION_PORT = 14
    ERR_ADDRESS_PARSING = 15
    ERR_MALFORMED_PACKET = 16
    ERR_SUBSCRIPTION_FAILURE = 17
    ERR_PAYLOAD_DECODE_FAILURE = 18
    ERR_COMPILE_DECODER = 19
    ERR_UNSUPPORTED_PACKET_TYPE = 20


def _get_mqtt_result(result_code: int) -> str:
    try:
        return MqttResultCode(result_code).name
    except ValueError:
        return 'UNKNOWN'


class MqttError(Exception):
    """A MQTT-specific error."""


class MqttClient:
    """A customized MQTT client.

    Attributes:
        client_id (str): A unique client_id.
        subscriptions (dict): A dictionary of subscriptions with qos and
            message ID properties
        on_message (callable): The callback when subscribed messages are
            received as `topic`(str), `message`(dict|str).
        is_connected (bool): Status of the connection to the broker.
        auto_connect (bool): Automatically attempts to connect when created
            or reconnect after disconnection.
        connect_retry_interval (int): Seconds between broker reconnect attempts.

    """
    def __init__(self,
                 client_id: str = __name__,
                 on_message: callable = None,
                 subscribe_default: 'str|list[str]' = None,
                 auto_connect: bool = True,
                 connect_retry_interval: int = 5,
                 **kwargs) -> None:
        """Initializes a managed MQTT client.
        
        Args:
            client_id (str): The client ID (default imports module `__name__`)
            on_message (callable): The callback when subscribed messages are
                received as `topic`(str), `message`(dict|str).
            subscribe_default (str|list[str]): The default subscription(s)
                established on re/connection.
            connect_retry_interval (int): Seconds between broker reconnect
                attempts if auto_connect is `True`.
            auto_connect (bool): Automatically attempts to connect when created
                or reconnect after disconnection.
        
        Keyword Args: 
            client_uid (str): defaults to True, appends a timestamp to the
                client_id to avoid being rejected by the host.
            bind_address (str): to bind to a specific IP (broken in Paho Python)
            on_connect (Callable): optional callback when connecting to broker
            on_disconnect (Callable): optional callback when disconnecting from
                broker
            on_log (Callable): optional callback for client logging
            host (str): override hostname
            port (int): defaults to 1883
            keepalive (int): defaults to 60 seconds
            username (str): override username
            password (str): override password
            ca_certs (str): path to the CA certificate
            certfile (str): path to the PEM certificate for the client
            keyfile (str): path to the PEM certificate for the client key
            qos (int): MQTT QoS 0=at most once (default), 1=at least once,
                2=exactly once
            thread_name (str): Optional tag to identify thread in logging.

        Raises:
            `MqttError` if the client_id is not valid.

        """
        if (str(os.getenv('DOCKER')).lower() in ['1', 'true']):
            dflt_host = 'fieldedge-broker'
        else:
            dflt_host = 'localhost'
        self._host: str = kwargs.get('host', os.getenv('MQTT_HOST', dflt_host))
        self._username: str = kwargs.get('username', os.getenv('MQTT_USER'))
        self._password = kwargs.get('password', os.getenv('MQTT_PASS'))
        self._port = int(kwargs.get('port', os.getenv('MQTT_PORT', '1883')))
        self._keepalive = int(kwargs.get('keepalive', 60))
        self._bind_address = kwargs.get('bind_address', '')
        self._ca_certs = kwargs.get('ca_certs', None)
        self._certfile = kwargs.get('certfile', None)
        self._keyfile = kwargs.get('keyfile', None)
        if not isinstance(client_id, str) or client_id == '':
            _log.error('Invalid client_id')
            raise MqttError('Invalid client_id')
        if not callable(on_message):
            _log.warning('No on_message specified')
        on_exit(self._cleanup)
        self.on_message = on_message
        self.on_connect: 'Callable|None' = kwargs.get('on_connect', None)
        self.on_disconnect: 'Callable|None' = kwargs.get('on_disconnect', None)
        self._qos = kwargs.get('qos', 0)
        self._thread_name: str = kwargs.get('thread_name', None)
        self._client_base_id = client_id
        self._client_id = None
        self._client_uid = kwargs.get('client_uid', True)
        if self._host.endswith('azure-devices.net'):
            _log.debug('Assuming %s is the Azure IoT Device ID', client_id)
            self._client_uid = False
        self.client_id = client_id
        self._clean_session = kwargs.get('clean_session', True)
        self._mqtt = PahoClient(clean_session=self._clean_session,
                                reconnect_on_failure=False)
        self._connect_timeout = 5
        self.connect_timeout = int(kwargs.get('connect_timeout', 5))
        self._subscriptions = {}
        self._connect_retry_interval: int = 0
        self.connect_retry_interval = connect_retry_interval
        self.auto_connect: bool = auto_connect
        self._failed_connect_attempts = 0
        if subscribe_default:
            if not isinstance(subscribe_default, list):
                subscribe_default = [subscribe_default]
            for sub in subscribe_default:
                self.subscribe(sub, self._qos)
        if self.auto_connect:
            self.connect()

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, uid: str):
        if not self._client_uid:
            self._client_id = uid
        else:
            if uid != self._client_base_id:
                if _vlog():
                    _log.debug('Updating client_id %s with new timestamp', uid)
                uid = self._client_base_id
            self._client_id = f'{uid}_{int(time())}'

    @property
    def is_connected(self) -> bool:
        return self._mqtt.is_connected()

    @property
    def subscriptions(self) -> 'dict[str, dict]':
        """The dictionary of subscriptions.
        
        Use subscribe or unsubscribe to change the dict.

        'topic' : { 'qos': (int), 'mid': (int) }

        """
        return self._subscriptions

    @property
    def failed_connection_attempts(self) -> int:
        return self._failed_connect_attempts

    @property
    def on_log(self) -> Callable:
        return self._mqtt.on_log

    @on_log.setter
    def on_log(self, callback: Callable):
        if not isinstance(callback, Callable):
            raise ValueError('Callback must be a function')
        self._mqtt.on_log = callback

    @property
    def connect_timeout(self) -> int:
        return int(self._mqtt._connect_timeout)

    @connect_timeout.setter
    def connect_timeout(self, value: 'int|float'):
        if (not isinstance(value, (int, float)) or
            not 0 < value <= 120):
            # invalid value
            raise ValueError('Connect timeout must be 1..120 seconds')
        self._connect_timeout = value
        self._mqtt._connect_timeout = float(value)

    @property
    def connect_retry_interval(self) -> int:
        return self._connect_retry_interval

    @connect_retry_interval.setter
    def connect_retry_interval(self, value: int):
        if not isinstance(value, int) or value < 0:
            raise ValueError('Retry interval must be integer 0 or higher')
        self._connect_retry_interval = value

    def _cleanup(self, *args):
        # TODO: logging raises an error since the log file was closed
        if _vlog():
            try:
                for arg in args:
                    _log.debug('mqtt cleanup called with arg = %s', arg)
                _log.debug('Terminating MQTT connection')
            except:
                pass
        self._mqtt.user_data_set('terminate')
        self._mqtt.loop_stop()
        self._mqtt.disconnect()

    def _unique_thread_name(self, before_names: 'list[str]') -> str:
        basename = 'MqttThread'
        if self._thread_name:
            basename += f'-{self._thread_name}'
        new_name = basename
        number = 1
        for name in before_names:
            if name.startswith(basename):
                number += 1
                new_name = f'{basename}-{number}'
        return new_name

    def connect(self):
        """Attempts to establish a connection to the broker and re-subscribe."""
        try:
            if _vlog():
                _log.debug('Attempting MQTT broker connection to %s as %s',
                           self._host, self.client_id)
            self._mqtt.reinitialise(client_id=self.client_id)
            self.connect_timeout = self._connect_timeout   #: just in case
            self._mqtt.user_data_set(None)
            self._mqtt.on_connect = self._mqtt_on_connect
            self._mqtt.on_disconnect = self._mqtt_on_disconnect
            self._mqtt.on_subscribe = self._mqtt_on_subscribe
            self._mqtt.on_message = self._mqtt_on_message
            if self._username or self._password:
                self._mqtt.username_pw_set(username=self._username,
                                           password=self._password)
            if self._port == 8883:
                self._mqtt.tls_set(ca_certs=self._ca_certs,
                                   certfile=self._certfile,
                                   keyfile=self._keyfile)
                # self._mqtt.tls_insecure_set(False)
            self._mqtt.connect(host=self._host,
                               port=self._port,
                               keepalive=self._keepalive,
                               bind_address=self._bind_address)
            threads_before = threading.enumerate()
            self._mqtt.loop_start()
            threads_after = threading.enumerate()
            new_thread = list(set(threads_after) - set(threads_before))[0]
            before_names = [t.name for t in threads_before]
            new_thread.name = self._unique_thread_name(before_names)
            _log.debug('New MQTT client thread: %s', new_thread.name)
            return
        except (ConnectionError, TimeoutError, gaierror, timeout) as exc:
            self._mqtt.loop_stop()
            self._failed_connect_attempts += 1
            _log.error('Failed attempt %d to connect to %s (%s)',
                       self._failed_connect_attempts, self._host, exc)
            if not self.auto_connect or self.connect_retry_interval <= 0:
                raise ConnectionError(str(exc)) from exc
            _log.debug('Retrying in %d seconds', self.connect_retry_interval)
            sleep(self.connect_retry_interval)
            self.connect()

    def disconnect(self):
        """Attempts to disconnect from the broker."""
        self._mqtt.user_data_set('terminate')
        self._mqtt.loop_stop()
        self._mqtt.disconnect()

    def _mqtt_on_connect(self,
                         client: PahoClient,
                         userdata: Any,
                         flags: dict,
                         result_code: int):
        """Internal callback re-subscribes on (re)connection."""
        self._failed_connect_attempts = 0
        if result_code == MqttResultCode.SUCCESS:
            if _vlog():
                _log.debug('Established MQTT connection to %s', self._host)
            for sub, meta in self.subscriptions.items():
                self._mqtt_subscribe(sub, qos=meta.get('qos', None))
            if callable(self.on_connect):
                self.on_connect(client, userdata, flags, result_code)
        else:
            _log.error('MQTT broker connection result code: %d (%s)',
                       result_code, _get_mqtt_result(result_code))

    def _mqtt_subscribe(self, topic: str, qos: int = 0):
        """Internal subscription handler assigns id indicating *subscribed*."""
        (result, mid) = self._mqtt.subscribe(topic=topic, qos=qos)
        if _vlog():
            _log.debug('%s subscribing to %s (qos=%d, mid=%d)',
                       self.client_id, topic, qos, mid)
        if result == MqttResultCode.SUCCESS:
            if mid == 0:
                _log.warning('Received mid=%d expected > 0', mid)
            self._subscriptions[topic]['mid'] = mid
        else:
            _log.error('MQTT Error %s subscribing to %s', result, topic)

    def subscribe(self, topic: str, qos: int = 0) -> None:
        """Adds a subscription.
        
        Subscriptions property is updated with qos and message id.
        Message id `mid` is 0 when not actively subscribed.

        Args:
            topic (str): The MQTT topic to subscribe to
            qos (int): The MQTT qos 0..2

        """
        if _vlog():
            _log.debug('Adding subscription %s (qos=%d)', topic, qos)
        self._subscriptions[topic] = {'qos': qos, 'mid': 0}
        if self.is_connected:
            self._mqtt_subscribe(topic, qos)
        else:
            _log.debug('MQTT not connected...subscribing to %s later', topic)

    def _mqtt_on_subscribe(self,
                           client: PahoClient,
                           userdata: Any,
                           mid: int,
                           granted_qos: 'tuple[int]'):
        match = ''
        for topic, detail in self.subscriptions.items():
            if mid == detail.get('mid', None):
                _log.info('Subscribed to %s (mid=%d, granted_qos=%s)',
                          topic, mid, granted_qos)
                match = topic
                break
        if not match:
            _log.error('Unable to match mid=%d to pending subscription', mid)

    def is_subscribed(self, topic: str) -> bool:
        """Returns True if the specified topic is an active subscription."""
        if (topic in self.subscriptions and
            self.subscriptions[topic]['mid'] > 0):
            return True
        return False

    def _mqtt_on_message(self,
                         client: PahoClient,
                         userdata: Any,
                         message: PahoMessage):
        """Internal callback on message simplifies passback to topic/payload."""
        payload = message.payload.decode()
        try:
            payload = json.loads(payload)
        except json.JSONDecodeError as exc:
            if _vlog():
                _log.debug('MQTT message payload non-JSON (%s)', exc)
        if _vlog():
            _log.debug('MQTT received message on topic "%s" (QoS=%d): "%s"',
                       message.topic, message.qos, payload)
            if userdata:
                _log.debug('MQTT client userdata: %s', userdata)
        self.on_message(message.topic, payload)

    def _mqtt_unsubscribe(self, topic: str):
        if _vlog():
            _log.debug('%s unsubscribing to %s', self.client_id, topic)
        (result, mid) = self._mqtt.unsubscribe(topic)
        if result != MqttResultCode.SUCCESS:
            _log.error('MQTT Error %s unsubscribing from %s (mid=%d)',
                       result, topic, mid)

    def unsubscribe(self, topic: str) -> None:
        """Removes a subscription.
        
        Args:
            topic (str): The MQTT topic to unsubscribe

        """
        if _vlog():
            _log.debug('Removing subscription %s', topic)
        if topic in self._subscriptions:
            del self._subscriptions[topic]
        if self.is_connected:
            self._mqtt_unsubscribe(topic)

    def _mqtt_on_disconnect(self,
                            client: PahoClient,
                            userdata: Any,
                            result_code: int):
        """Internal callback when disconnected, clears subscription status."""
        for subscription in self.subscriptions.values():
            subscription['mid'] = 0
        if userdata != 'terminate':
            _log.warning('MQTT broker disconnected - result code %d (%s)',
                         result_code, _get_mqtt_result(result_code))
            # reconnect handling is managed automatically by Paho library
        if callable(self.on_disconnect):
            self.on_disconnect(client, userdata, result_code)

    def publish(self,
                topic: str,
                message: 'str|dict|None',
                qos: int = 1,
                camel_keys: bool = False,
                wait_for_publish: float = None,
                ) -> bool:
        """Publishes a message to a MQTT topic.

        If the message is a dictionary, validates JSON compatibility.
        
        Args:
            topic: The MQTT topic
            message: The message payload
            qos: The MQTT Quality of Service (0, 1 or 2)
            camel_keys: Ensures message dictionary keys are JSON camelCase style
            wait_for_publish: If present waits up to this timeout to complete

        Returns:
            True if successful, else False.

        """
        if message and not isinstance(message, (str, dict)):
            raise ValueError(f'Invalid message {message}')
        if self._host.endswith('.azure-devices.net'):
            device_to_cloud = f'devices/{self.client_id}/messages/events/'
            if device_to_cloud not in topic:
                _log.warning('Applying Azure device-to-cloud topic prefix')
                topic = f'{device_to_cloud}{topic}'
        if isinstance(message, dict):
            message = json.dumps(json_compatible(message, camel_keys),
                                 skipkeys=True)
        if not isinstance(qos, int) or qos not in range(0, 3):
            _log.warning('Invalid MQTT QoS %s - using QoS 1', qos)
            qos = 1
        publish_info = self._mqtt.publish(topic=topic, payload=message, qos=qos)
        if wait_for_publish:
            try:
                publish_info.wait_for_publish(wait_for_publish)
            except (ValueError, RuntimeError) as exc:
                _log.error("Wait error: %s", exc)
        if publish_info.rc != MqttResultCode.SUCCESS:
            _log.error('Publishing error %d (%s)',
                       publish_info.rc, _get_mqtt_result(publish_info.rc))
            return False
        if _vlog():
            _log.debug('MQTT published (mid=%d, qos=%d) %s: %s',
                       publish_info.mid, qos, topic, message)
        return True


def _vlog() -> bool:
    return verbose_logging('mqtt')



"""Message management utilities for FieldEdge applications."""

import base64
from dataclasses import dataclass, field
from enum import IntEnum

__all__ = ['MessageState', 'MessageMeta', 'MessageStore']


class MessageState(IntEnum):
    """Message state enumerated type."""
    UNAVAILABLE = 0
    RX_PENDING = 1
    RX_COMPLETE = 2
    RX_RETRIEVED = 3 
    TX_QUEUED = 4
    TX_SENDING = 5
    TX_COMPLETE = 6
    TX_FAILED = 7
    TX_CANCELLED = 8


@dataclass
class MessageMeta:
    """Message metadata."""
    id: 'int|str'
    mo: bool   # mo = Mobile-Originated (else Mobile-Terminated)
    data_b64: str = ''   # Base64-encoded string
    
    @property
    def size(self) -> int:
        if not isinstance(self.data_b64, str) or len(self.data_b64) == 0:
            return 0
        return len(base64.b64decode(self.data_b64))


@dataclass
class MessageStore:
    """A temporary storage buffer in memory for messages."""
    tx_queue: 'list[MessageMeta]' = field(default_factory=list)
    rx_queue: 'list[MessageMeta]' = field(default_factory=list)
    byte_count: int = 0
    last_mo_id: int = 0
    last_mt_id: int = 0

    def add(self, message: 'MessageMeta') -> None:
        """Adds a message to the buffer."""
        if not isinstance(message, MessageMeta):
            raise ValueError('Invalid message metadata')
        queue = self.tx_queue if message.mo else self.rx_queue
        if any(queued.id == message.id for queued in queue):
            raise ValueError(f'Duplicate id {message.id} found')
        if queue == self.tx_queue:
            self.last_mo_id = message.id
        else:
            self.last_mt_id = message.id
        queue.append(message)
        self.byte_count += message.size

    def get(self, id: int, mo: bool = True, retain: bool = False) -> MessageMeta:
        """Retrieves a message from the buffer, optionally retaining it.
        
        id -1 indicates the first enqueued message.
        """
        queue = self.tx_queue if mo else self.rx_queue
        if id == -1 and len(queue) > 0:
            if not retain:
                return queue.pop(0)
            return queue[0]
        for i, message in enumerate(queue):
            if message.id == id:
                if not retain:
                    return queue.pop(i)
                return message
        raise ValueError(f'Message {id} not found in queue')



"""A threaded timer class that allows flexible reconfiguration.

"""
import logging
import threading
from time import time
from typing import Callable

from .logger import verbose_logging

_log = logging.getLogger(__name__)


class RepeatingTimer(threading.Thread):
    """A repeating thread to call a function, can be stopped/restarted/changed.
    
    Embedded tasks can use threading for continuous repeated operations.  A
    *RepeatingTimer* can be started, stopped, restarted and reconfigured.

    A Thread that counts down seconds using sleep increments, then calls back 
    to a function with any provided arguments.
    Optional auto_start feature starts the thread and the timer, in this case 
    the user doesn't need to explicitly start() then start_timer().

    Attributes:
        name (str): An optional descriptive name for the Thread.
        interval (int): Repeating timer interval in seconds (0=disabled).
        sleep_chunk (float): The fraction of seconds between processing ticks.
        max_drift (int): Delay allowed to impact the next countdown after
            running the target function.
        defer (bool): Waits until the first interval before triggering the
            target function (default = True)

    """
    def __init__(self,
                 seconds: int,
                 target: Callable,
                 args: tuple = None,
                 kwargs: dict = None,
                 name: str = None,
                 sleep_chunk: float = 0.25,
                 max_drift: int = None,
                 auto_start: bool = False,
                 defer: bool = True,
                 daemon: bool = True,
                 ):
        """Sets up a RepeatingTimer thread.

        Args:
            seconds: Interval for timer repeat.
            target: The function to execute each timer expiry.
            args: Positional arguments required by the target.
            kwargs: Optional keyword arguments to pass into the target.
            name: Optional thread name.
            sleep_chunk: Tick seconds between expiry checks.
            max_drift: Number of seconds delay from function call, to tolerate.
            auto_start: Starts the thread and timer when created.
            defer: Set if first target waits for timer expiry.
            daemon: Set if thread is a daemon (default)

        Raises:
            ValueError if seconds is not an integer.
        """
        if not (isinstance(seconds, int) and seconds >= 0):
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            raise ValueError(err_str)
        super().__init__(daemon=daemon)
        self.name = name or f'{target.__name__}_timer_thread'
        self._interval: int = 0
        self.interval = seconds
        if target is None:
            _log.warning('No target specified for RepeatingTimer %s', self.name)
        self.target = target
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.sleep_chunk = sleep_chunk
        self.defer = defer
        self._terminate_event = threading.Event()
        self._start_event = threading.Event()
        self._reset_event = threading.Event()
        self._count = self.interval / self.sleep_chunk
        self._timesync: int = None
        self._max_drift = None
        self.max_drift = max_drift
        if auto_start:
            self.start()
            self.start_timer()

    @property
    def interval(self) -> int:
        return self._interval

    @interval.setter
    def interval(self, val: int):
        if val is None or not isinstance(val, int) or val < 0:
            raise ValueError('Interval must be a positive integer or zero')
        self._interval = val

    @property
    def sleep_chunk(self) -> float:
        return self._sleep_chunk

    @sleep_chunk.setter
    def sleep_chunk(self, value: float):
        if 1 % value != 0:
            raise ValueError('sleep_chunk must evenly divide 1 second')
        self._sleep_chunk = value

    @property
    def is_running(self) -> bool:
        return self._start_event.is_set()

    @property
    def max_drift(self) -> 'int|None':
        return self._max_drift

    @max_drift.setter
    def max_drift(self, val: 'int|None'):
        if (val is not None and
            (not isinstance(val, int) or val < 0 or val >= self.interval)):
            raise ValueError('max_drift must be None, 0'
                             ' or integer < interval')
        self._max_drift = val

    def _resync(self) -> int:
        """Used to adjust the next countdown to account for drift."""
        if self.max_drift is not None:
            drift = (int(time()) - self._timesync) % self.interval
            if drift > self.max_drift:
                _log.debug('Compensating for drift of %d seconds', drift)
                return drift
        return 0

    def run(self):
        """*Note: runs automatically, not meant to be called explicitly.*
        
        Counts down the interval, checking every ``sleep_chunk`` for expiry.
        """
        while not self._terminate_event.is_set():
            while (self._count > 0
                   and self._start_event.is_set()
                   and self.interval > 0):
                if _vlog():
                    if (self._count * self.sleep_chunk
                        - int(self._count * self.sleep_chunk)
                        <= 0.0):
                        #: log debug message at reasonable interval
                        _log.debug('%s countdown: %d (%d s) @ step %0.2f',
                                   self.name, self._count, self.interval,
                                   self.sleep_chunk)
                if self._reset_event.wait(self.sleep_chunk):
                    # reset -> restart the countdown
                    self._reset_event.clear()
                    self._count = self.interval / self.sleep_chunk
                self._count -= 1
                if self._count <= 0:
                    try:   # countdown expired, trigger function and restart
                        self.target(*self.args, **self.kwargs)
                        drift_adjusted = self.interval - self._resync()
                        self._count = drift_adjusted / self.sleep_chunk
                    except BaseException as exc:
                        _log.error('Exception in %s: %s', self.name, exc)
                        raise

    def start_timer(self):
        """Initially start the repeating timer."""
        self._timesync = int(time())
        self._start_event.set()
        if self.interval > 0:
            _log.info('%s timer started (%d s)', self.name, self.interval)
            if not self.defer:
                _log.debug('Triggering initial call to %s',
                           self.target.__name__)
                self.target(*self.args, **self.kwargs)
        else:
            _log.warning('%s timer cannot trigger (interval=0)', self.name)

    def stop_timer(self):
        """Stop the repeating timer."""
        self._start_event.clear()
        _log.info('%s timer stopped (%d s)', self.name, self.interval)
        self._count = self.interval / self.sleep_chunk

    def restart_timer(self, trigger_immediate: bool = None):
        """Restart the repeating timer (after an interval change)."""
        if trigger_immediate is None:
            trigger_immediate = not self.defer
        if self._start_event.is_set():
            self._reset_event.set()
        else:
            self._start_event.set()
        if self.interval > 0:
            _log.info('%s timer restarted (%d s)', self.name, self.interval)
            if trigger_immediate:
                self.target(*self.args, **self.kwargs)
        else:
            _log.warning('%s timer cannot trigger (interval=0)', self.name)

    def change_interval(self, seconds: int, trigger_immediate: bool = None):
        """Change the timer interval and restart it.
        
        Args:
            seconds (int): The new interval in seconds.
        
        Raises:
            ValueError if seconds is not an integer.

        """
        if trigger_immediate is None:
            trigger_immediate = not self.defer
        if (isinstance(seconds, int) and seconds >= 0):
            _log.info('%s timer interval changed (old: %d s new: %d s)',
                      self.name, self.interval, seconds)
            self.interval = seconds
            self._count = self.interval / self.sleep_chunk
            self.restart_timer(trigger_immediate)
        else:
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            _log.error(err_str)
            raise ValueError(err_str)

    def terminate(self):
        """Terminate the timer. (Cannot be restarted)"""
        self.stop_timer()
        self._terminate_event.set()
        _log.info('%s timer terminated', self.name)

    def join(self, timeout=None):
        super().join(timeout)
        return self.target


def _vlog() -> bool:
    return verbose_logging('timer')
